program Font3D;

uses
  Windows, Messages, OpenGL, Graphics;

const
  WND_TITLE       = '3D Font App by Arno van der Vegt';
  FPS_TIMER       = 1;                                             // Timer to calculate FPS
  FPS_INTERVAL    = 1000;                                          // Calculate FPS every 1000 ms

  gGlobalAmbient  : array[0..3] of glFloat = (1.0, 1.0, 1.0, 1.0); // Set Ambient Lighting

  gLight0Pos      : array[0..3] of glFloat = (0.0, 5.0, 0.0, 1.0); // Position for Light0
  gLight0Ambient  : array[0..3] of glFloat = (0.2, 0.2, 0.2, 1.0); // Ambient setting for light0
  gLight0Diffuse  : array[0..3] of glFloat = (1.0, 1.0, 1.0, 1.0); // Diffuse setting for Light0
  gLight0Specular : array[0..3] of glFloat = (0.8, 0.8, 0.8, 1.0); // Specular lighting for Light0

  gLModelAmbient  : array[0..3] of glFloat = (1.0, 1.0, 1.0, 1.0); // And More Ambient Light

type
  TFontSize       = record
                      fBoxX, fBoxY : single;
                    end;

  TRGB            = array[0..2] of byte;

var
  h_Wnd           : HWND;                           // Global window handle
  h_DC            : HDC;                            // Global device context
  h_RC            : HGLRC;                          // OpenGL rendering context

  gKeys           : array[0..255] of boolean;       // Holds keystrokes
  gFPSCount       : integer = 0;                    // Counter for FPS
  gElapsedTime    : cardinal;                       // Elapsed time between frames

  gFontSizes      : array[0..1023] of TFontSize;    // Buffer for character size information
  gFontList       : integer;                        // Handle to character display lists
  gTexture        : array[0..127, 0..127] of TRGB;  // Texture data
  gTextureHandle  : integer;                        // Handle to OpenGL texture

{$R *.RES}

{------------------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)              }
{------------------------------------------------------------------------------}
function IntToStr(pNum: integer): string; // Using SysUtils increase file size by 100K
begin
  Str(pNum, Result);
end; { IntToStr }

{------------------------------------------------------------------------------}
{ Function to make a texture from the pixel data                               }
{------------------------------------------------------------------------------}
procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;
procedure glGenTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;

function CreateTexture(pWidth, pHeight: integer; pData: pointer): glUint;
begin
  glGenTextures(1, Result);
  glBindTexture(GL_TEXTURE_2D, Result);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, pWidth, pHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, pData);
end; { CreateTexture }

{------------------------------------------------------------------------------}
{ Function to draw the 3d font                                                 }
{------------------------------------------------------------------------------}
procedure glWrite3D(pText: string; pX, pY, pZ: single; pCenter: boolean);
var
  i  : integer;
  lX : single;
begin
  // Check empty string...
  if (pText = '') then Exit;

  lX:= pX;

  if pCenter then
  begin
    // Calculate the half width of the text...
    for i:= 1 to Length(pText) do lX:= lX - gFontSizes[Ord(pText[i])].fBoxX;
  end;

  glPushMatrix();

    // Center the text...
    glTranslatef(lX, pY, pZ);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, gTextureHandle);

    // Enable texture genarate...
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);

      // Set the texture genarate mode, creates the texture coordinates...
      glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
      glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

      // Set the font list...
      glListBase(gFontList);

      glCallLists(Length(pText),      // Length of the string
                  GL_UNSIGNED_BYTE,   // Chars, one byte each
                  @pText[1]);         // Pointer to the first char of the string

    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);

  glPopMatrix();
end; { glWrite3D }

{------------------------------------------------------------------------------}
{ Function to draw the actual scene                                            }
{------------------------------------------------------------------------------}
procedure glDraw();
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);          // Clear The Screen And The Depth Buffer
  glLoadIdentity();                                             // Reset The View

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @gGlobalAmbient);      // Set The Global Ambient Light Model

  glLightfv(GL_LIGHT0, GL_POSITION, @gLight0Pos);               // Set The Lights Position
  glLightfv(GL_LIGHT0, GL_AMBIENT,  @gLight0Ambient);           // Set The Ambient Light
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  @gLight0Diffuse);           // Set The Diffuse Light
  glLightfv(GL_LIGHT0, GL_SPECULAR, @gLight0Specular);          // Set Up Specular Lighting

  glEnable(GL_LIGHTING);                                        // Enable Lighting
  glEnable(GL_LIGHT0);                                          // Enable Light0

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @gLModelAmbient);      // Set The Ambient Light Model

  glTranslatef(0.0, -0.3, -4.0);

  glRotatef(Sin(GetTickCount() / 1000) * 60.0, 0.0, 1.0, 0.0);  // Rotate arount Y

  glWrite3D('AbCd',  0.5, -0.5, 0.0, true);
  glWrite3D('AbCd', -0.5,  0.5, 0.0, true);
end; { glDraw }

{------------------------------------------------------------------------------}
{  Initialise OpenGL                                                           }
{------------------------------------------------------------------------------}
procedure glInit();
const
  lColor : array[0..1] of TRGB = ((255, 255, 255), (255, 0, 0));
var
  i, j : integer;
begin
  glClearColor(0.0, 0.0, 0.0, 0.0); 	                  // Black Background
  glShadeModel(GL_SMOOTH);                              // Enables Smooth Color Shading
  glClearDepth(1.0);                                    // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);                              // Enable Depth Buffer
  glDepthFunc(GL_LESS);		                              // The Type Of Depth Test To Do
  glEnable(GL_TEXTURE_2D);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);    // Realy Nice perspective calculations

  // Create a texture with 16 red and white rectangles...
  for i:= 0 to 127 do for j:= 0 to 127 do gTexture[i, j]:= lColor[(i shr 5 + j shr 5) and 1];

  gTextureHandle:= CreateTexture(128, 128, @gTexture);
end; { glInit }

{------------------------------------------------------------------------------}
{  Handle window resize                                                        }
{------------------------------------------------------------------------------}
procedure glResizeWnd(pWidth, pHeight: integer);
begin
  if (pHeight = 0) then pHeight:= 1;                   // Prevent divide by zero exception
  glViewport(0, 0, pWidth, pHeight);                   // Set the viewport for the OpenGL window
  glMatrixMode(GL_PROJECTION);                         // Change Matrix Mode to Projection
  glLoadIdentity();                                    // Reset View
  gluPerspective(45.0, pWidth / pHeight, 1.0, 100.0);  // Do the perspective calculations. Last value = max clipping depth

  glMatrixMode(GL_MODELVIEW);                          // Return to the modelview matrix
  glLoadIdentity();                                    // Reset View
end; { glResizeWnd }

{------------------------------------------------------------------------------}
{ Processes all the keystrokes                                                 }
{------------------------------------------------------------------------------}
procedure ProcessKeys();
begin
  if gKeys[32] then
  begin
    gKeys[32]:= false;
  end;
end; { ProcessKeys }

{------------------------------------------------------------------------------}
{  Determines the applications response to the messages received              }
{------------------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  Result:= 0;

  case (Msg) of
    WM_CREATE:
    begin
                       // Insert stuff you want executed when the program starts
    end;
    WM_CLOSE:
    begin
      PostQuitMessage(0);
    end;
    WM_KEYDOWN:        // Set the pressed key (wparam) to equal true so we can check if its pressed
    begin
      gKeys[wParam]:= true;
    end;
    WM_KEYUP:          // Set the released key (wparam) to equal false so we can check if its pressed
    begin
      gKeys[wParam]:= false;
    end;
    WM_SIZE:           // Resize the window with the new width and height
    begin
      glResizeWnd(LOWORD(lParam),HIWORD(lParam));
    end;
    WM_MOUSEMOVE:
    begin
    end;
    WM_TIMER:          // Add code here for all timers to be used.
    begin
      if (wParam = FPS_TIMER) then
      begin
          // Calculate to get per Second incase intercal is less or greater than 1 second
        gFPSCount:= Round(gFPSCount * 1000 / FPS_INTERVAL);
        SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + intToStr(gFPSCount) + ' FPS]'));
        gFPSCount:= 0;
      end;
    end else           // Default result if nothing happens
    begin
      Result:= DefWindowProc(hWnd, Msg, wParam, lParam);
    end;
  end;
end; { WndProc }

{------------------------------------------------------------------------------}
{ Properly destroys the window created at startup (no memory leaks)            }
{------------------------------------------------------------------------------}
procedure glKillWnd(Fullscreen: boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if not wglMakeCurrent(h_DC, 0) then
  begin
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);
  end;

  // Attempts to delete the rendering context
  if not wglDeleteContext(h_RC) then
  begin
    MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_RC:= 0;
  end;

  // Attemps to release the device context
  if ((h_DC = 1) and (ReleaseDC(h_Wnd, h_DC) <> 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC:= 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd:= 0;
  end;

  // Attempts to unregister the window class
  if not UnRegisterClass('OpenGL', hInstance) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance:= 0;
  end;
end; { glKillWnd }

procedure glCreateFont();
var
  lFont : TFont;
begin
  lFont:= TFont.Create();

  lFont.Name:= 'Arial';                  // Set the font name for the windows font

  SelectObject(h_DC, lFont.Handle);      // Select font device context

  // Create display lists for each character in the font:
  gFontList:= glGenLists(256);
  wglUseFontOutlines(h_DC,               // Device context of font -source-
                     0,                  // First character
                     256,                // Number of characters
                     gFontList,          // Handle of font display lists
                     0.0,                // This is the sampling tolerance. Higher values create less detailed outlines.
                     0.2,                // This is the extrusion depth.
                     WGL_FONT_POLYGONS,  // What kind of output? You can also use WGL_FONT_LINES
                     @gFontSizes);       // Array to store info about character sizes

  lFont.Free();
end; { glCreateFont }

{------------------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it            }
{------------------------------------------------------------------------------}
function glCreateWnd(Width, Height: integer; Fullscreen: boolean; PixelDepth: integer): boolean;
var
  wndClass         : TWndClass;               // Window class
  dwStyle          : DWORD;                   // Window styles
  dwExStyle        : DWORD;                   // Extended window styles
  dmScreenSettings : DEVMODE;                 // Screen settings (fullscreen, etc...)
  PixelFormat      : GLuint;                  // Settings for the OpenGL rendering
  h_Instance       : HINST;                   // Current instance
  pfd              : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
begin
  h_Instance:= GetModuleHandle(nil);          // Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));    // Clear the window class structure

  with wndClass do                            // Set up the window class
  begin
    style         := CS_HREDRAW or            // Redraws entire window if length changes
                     CS_VREDRAW or            // Redraws entire window if height changes
                     CS_OWNDC;                // Unique device context for the window
    lpfnWndProc   := @WndProc;                // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do
    begin                                      // Set parameters for the screen setting
      dmSize      := SizeOf(dmScreenSettings);
      dmPelsWidth := Width;                    // Window width
      dmPelsHeight:= Height;                   // Window height
      dmBitsPerPel:= PixelDepth;               // Window color depth
      dmFields    := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen:= false;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle  := WS_POPUP        or        // Creates a popup window
                WS_CLIPCHILDREN or        // Doesn't draw within child windows
                WS_CLIPSIBLINGS;          // Doesn't draw within sibling windows
    dwExStyle:= WS_EX_APPWINDOW;          // Top level window
    ShowCursor(false);                    // Turn of the cursor (gets in the way)
  end else
  begin
    dwStyle  := WS_OVERLAPPEDWINDOW or    // Creates an overlapping window
                WS_CLIPCHILDREN     or    // Doesn't draw within child windows
                WS_CLIPSIBLINGS;          // Doesn't draw within sibling windows
    dwExStyle:= WS_EX_APPWINDOW     or    // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd:= CreateWindowEx(dwExStyle,       // Extended window styles
                         'OpenGL',        // Class name
                         WND_TITLE,       // Window title (caption)
                         dwStyle,         // Window styles
                         0, 0,            // Window position
                         Width, Height,   // Size of window
                         0,               // No parent window
                         0,               // No menu
                         h_Instance,      // Instance
                         nil);            // Pass nothing to WM_CREATE
  if (h_Wnd = 0) then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Try to get a device context
  h_DC:= GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                              // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW or           // Buffer supports drawing to window
                       PFD_SUPPORT_OPENGL or           // Buffer supports OpenGL drawing
                       PFD_DOUBLEBUFFER;               // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;                  // RGBA color format
    cColorBits      := PixelDepth;                     // OpenGL color depth
    cRedBits        := 0;                              // Number of red bitplanes
    cRedShift       := 0;                              // Shift count for red bitplanes
    cGreenBits      := 0;                              // Number of green bitplanes
    cGreenShift     := 0;                              // Shift count for green bitplanes
    cBlueBits       := 0;                              // Number of blue bitplanes
    cBlueShift      := 0;                              // Shift count for blue bitplanes
    cAlphaBits      := 0;                              // Not supported
    cAlphaShift     := 0;                              // Not supported
    cAccumBits      := 0;                              // No accumulation buffer
    cAccumRedBits   := 0;                              // Number of red bits in a-buffer
    cAccumGreenBits := 0;                              // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                              // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                              // Number of alpha bits in a-buffer
    cDepthBits      := 16;                             // Specifies the depth of the depth buffer
    cStencilBits    := 0;                              // Turn off stencil buffer
    cAuxBuffers     := 0;                              // Not supported
    iLayerType      := PFD_MAIN_PLANE;                 // Ignored
    bReserved       := 0;                              // Number of overlay and underlay planes
    dwLayerMask     := 0;                              // Ignored
    dwVisibleMask   := 0;                              // Transparent color of underlay plane
    dwDamageMask    := 0;                              // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat:= ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC:= wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if not wglMakeCurrent(h_DC, h_RC) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result:= false;
    Exit;
  end;

  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  // Create the 3d font
  glCreateFont();

  Result:= true;
end; { glCreateWnd }

{------------------------------------------------------------------------------}
{ Main message loop for the application                                        }
{------------------------------------------------------------------------------}
function WinMain(hInstance: HINST; hPrevInstance: HINST;
                   lpCmdLine: PChar; nCmdShow: integer): integer; stdcall;
var
  lMsg                  : TMsg;
  lFinished             : boolean;
  lDemoStart, lLastTime : DWord;
begin
  lFinished:= false;

  if not glCreateWnd(800, 600, FALSE, 32) then         // Perform application initialization:
  begin
    Result:= 0;
    Exit;
  end;

  lDemoStart:= GetTickCount();                         // Get Time when demo started

  // Main message loop:
  while not lFinished do
  begin
    if PeekMessage(lMsg, 0, 0, 0, PM_REMOVE) then      // Check if there is a message for this window
    begin
      if (lMsg.message = WM_QUIT) then                 // If WM_QUIT message received then we are done
      begin
        lFinished:= true
      end else
      begin                                            // Else translate and dispatch the message to this window
        TranslateMessage(lMsg);
        DispatchMessage(lMsg);
      end;
    end else
    begin
      Inc(gFPSCount);                                  // Increment FPS Counter

      lLastTime   := gElapsedTime;
      gElapsedTime:= GetTickCount() - lDemoStart;      // Calculate Elapsed Time
      gElapsedTime:= (lLastTime + gElapsedTime) div 2; // Average it out for smoother movement

      glDraw();                                        // Draw the scene
      SwapBuffers(h_DC);                               // Display the scene

      if gKeys[VK_ESCAPE] then                         // If user pressed ESC then set finised TRUE
      begin
        lFinished:= true
      end else
      begin
        ProcessKeys();                                 // Check for any other key Pressed
      end;
    end;
  end;
  glKillWnd(false);
  Result:= lMsg.wParam;
end; { WinMain }

begin
  WinMain(hInstance, hPrevInst, CmdLine, CmdShow);
end.
